/*
  ****************************************
  *         L E D       C L O C K        *
  ****************************************
  *                                      *
  *This is the firmware of the LED CLOCK *
  *by Leonardo Miliani                   *
  *                                      *
  *The schematics and the last revision  *
  *of this software can be found at      *
  *www DOT leonardomiliani DOT com       *
  *                                      *
  ****************************************
  *Released under the therms of the      *
  *GNU General Public Licence (GPL) 3.0  *
  *or newer.                             *
  ****************************************
  *           Copyleft 2013              *
  ****************************************
  
  
  
  ****************************************
  Tested microcontrollers:
  Atmega328P & Atmega168P @ 8MHz
  
  Fuses:
  Low fuse: $E2
  High fuse: $D7
  Ext fuse: $07
  
  No external crystal needed
  ****************************************
  
  Version: 1.5.3
  ****************************************
  
*/

/**************************
  include required libraries
***************************/
//lib to monitor PCINTs
#define NO_PORTB_PINCHANGES 
#define NO_PORTC_PINCHANGES 
#define NO_PIN_STATE
#define DISABLE_PCINT_MULTI_SERVICE
#include "PinChangeInt.h"
//libs to access RTC
#include <Wire.h>
#include <Rtc_Pcf8563.h>
//schedulers
#include "leOS2.h"
#include "looper.h"
//needed to access internal EEPROM
#include "EEPROM.h"

/**************************
  instantiate the classes
***************************/
Rtc_Pcf8563 rtc;
leOS2 myOS;
looper myScheduler;

/**************************
  global variables
***************************/
volatile byte tHours, tMinutes, tSeconds; //used to keep the time locally
volatile byte FOLLOW_AMB_LIGHT = 0; //if set, manage the luminosity of the leds must vary depending on ambient light
byte NIGHT_MODE = 0; //if set, light off the clock during the night (00:00 - 06:00)
boolean nightModeOn = false;
volatile const byte ON = 0; //used to light on the LEDs
volatile const byte OFF = 255; //used to light off the LEDs
volatile boolean timeChanged = false; //used to inform the main loop that the clock must be updated
byte ledMatrix[5] = {0, 0, 0, 0, 0}; //used to store the status of the S/R pins depending of the time to show
volatile byte currentLedBrightness = 255; //used to fade the brightness of the LEDs
volatile byte readLedBrightness = 255; //used to fade the brightness of the LEDs
const unsigned int EEPROM_SETTINGS_ADDRESS = 100; //starting address of the EEPROM cells that will contain the settings of the clock

/**************************
  pin declaration
***************************/
const byte PIN_DATA = 0; //S/R data pin
const byte PIN_CLOCK = 1; //S/R clock pin
//const byte RTC_INT = 2; //this can be used for RTC interrupt with INT0
const byte PIN_LATCH = 3; //S/R latch pin
const byte RTC_CLK = 4; //used to get the clock from the RTC that updates the local time
volatile const byte LED_POWER = 5; //pin that drives the transistor that powers the LEDs
const byte btnSET = 6; //button SET/ADJUST
const byte btnINC = 7; //button INC/AMB_LIGHT
const byte btnDEC = 8; //button DEC/NIGHT_MODE
volatile const byte AMB_LIGHT = A0; //photoresistor


//setup
void setup() {
    //set up the scheduler
    myOS.begin(5000); //5s timeout to avoid tasks' crashes
    myOS.addTask(readAmbLight, 62, PAUSED);
    //myOS.addTask(readTime, 225000); //every ~1 hour, read the correct time from RTC
    myOS.addTask(blinkClock, 15, PAUSED); //LEDs blinking when programming
    myOS.addTask(writeAmbLight, 6, PAUSED); //fade the brightness of the LEDs
    myScheduler.addJob(readTime, 3600000UL); //every ~1 hour, read the correct time from RTC
    
    //LEDs off at startup
    analogWrite(LED_POWER, OFF);
    
    //set pins to drive the clock
    pinMode(PIN_DATA, OUTPUT);
    pinMode(PIN_CLOCK, OUTPUT);
    pinMode(PIN_LATCH, OUTPUT);
    //set pin buttons as input w/pullup
    pinMode(btnSET, INPUT_PULLUP);
    pinMode(btnINC, INPUT_PULLUP);
    pinMode(btnDEC, INPUT_PULLUP);

    delay(100);
    
    //hello world!
    lightAll(1);
    pushData(ledMatrix);
    
    //check if the user wants to run a test
    //test1?
    if (buttonPressed(btnDEC)) {
        test1();
    }
    
    //test 2?
    if (buttonPressed(btnINC)) {
        test2();
    }  

    //check if the RTC is running
    rtc.getDate();
    if ((rtc.getDay() == 0) && (rtc.getMonth() == 0)) { //RTC has lost its data
        rtc.initClock(); //initialize the RTC
        rtc.setDate(1, 2, 1, 0, 13); // set to 1/1/2013 - usesell because LED CLOCK will only show time
        rtc.setTime(0, 0, 0); //set to 00:00:00
    }
    //load the current time in local variables
    rtc.getTime();
    tSeconds = rtc.getSecond();
    tMinutes = rtc.getMinute();
    tHours = rtc.getHour();
    
    //attach a PCINT to the RTC CLK_OUT pin
    pinMode(RTC_CLK, INPUT_PULLUP);
    PCintPort::attachInterrupt(RTC_CLK, &incrementTime, FALLING);
    rtc.setSquareWave(SQW_1HZ); //1 Hz -> 1 sec.
    
    //load settings from EEPROM
    loadSettings();
    
    //LEDs on
    if (!FOLLOW_AMB_LIGHT) {
        analogWrite(LED_POWER, ON);
    } else {
        myOS.restartTask(readAmbLight);
        myOS.restartTask(writeAmbLight);
    }
}


//main loop
void loop() {
    updateClock(); //update the LED matrix
    
    //check if the user pressed the SET/ADJUST button
    if (!digitalRead(btnSET)) {
        if (waitButtonPress(btnSET)) {
            programClock(); //program the clock
        }
    }
    
    //check if the user pressed the INC/AMB_LIGHT button
    if (!digitalRead(btnINC)) {
        if (waitButtonPress(btnINC)) {
            //do changes only when not in night mode
            if (!nightModeOn) {
                FOLLOW_AMB_LIGHT ^= 1; //modify the option that tells the clock to follow the amb light or not
                saveSettings(); //store new value into EEPROM
                if (FOLLOW_AMB_LIGHT) {
                    myOS.restartTask(readAmbLight);
                    myOS.restartTask(writeAmbLight);
                } else {
                    myOS.pauseTask(readAmbLight);
                    myOS.pauseTask(writeAmbLight);
                    analogWrite(LED_POWER, ON);
                }
                waitForRelease(btnINC); //wait for button release
            }
        }
    }
    
    //check if the user pressed the DEC/NIGHT_MODE button
    if (!digitalRead(btnDEC)) {
        if (waitButtonPress(btnDEC)) {
            NIGHT_MODE ^= 1; //modify the option that tells the clock to follow the amb light or not
            saveSettings(); //store new value into EEPROM
            nightModeOn = false;
            if (NIGHT_MODE) {
                lightAll(2);
            } else {
                lightAll(1);
            }
            waitForRelease(btnDEC); //wait for button release
        }
    }
    myScheduler.scheduler(); //do the jobs
}



//wait for button release
inline void waitForRelease(byte tempBtn) {
    while (!digitalRead(tempBtn)) { //wait for button release
        updateClock();
    }
}

//wait for a button pressed 1.5 seconds
byte waitButtonPress(byte tempBtn) {
    //read 10 times the buttons - exit if button if released before 1 second
    byte pressed = 10;
    do {
        delay(100);
        updateClock();
        if (!digitalRead(tempBtn)) {
            pressed--;
        } else {
            break;
        }
    } while (pressed > 0);
    return (!pressed);
}


//light on all the LEDs
void lightAll(byte loops) {
    byte secMatrix[5] = {255, 255, 255, 255, 255}; //all LEDs on
    
    //LEDs steady on
    if (FOLLOW_AMB_LIGHT) {
        myOS.pauseTask(readAmbLight);
        myOS.pauseTask(writeAmbLight);
    }

    pushData(secMatrix);
    do {
        analogWrite(LED_POWER, ON);
        delay(300);
        analogWrite(LED_POWER, OFF);
        delay(200);
        loops--;
    } while (loops);
    
    updateTime(tSeconds, tMinutes, tHours);
    pushData(ledMatrix);
    if (FOLLOW_AMB_LIGHT) {
        myOS.restartTask(readAmbLight);
        myOS.restartTask(writeAmbLight);
    } else {
        analogWrite(LED_POWER, ON);
    }
}


void programClock() {
    byte tempSeconds = tSeconds;
    byte tempMinutes = tMinutes;
    byte tempHours = tHours;
    
    //flash to alert the user to release the button
    while (!digitalRead(btnSET)) {
        lightAll(1);
        updateClock();
    }
    
    //if follow amb light mode is on, I disable it
    if (FOLLOW_AMB_LIGHT) {
        myOS.pauseTask(readAmbLight);
        myOS.pauseTask(writeAmbLight);
    }  
    myOS.restartTask(blinkClock); //start LEDs blinking
    // ** change the time
    updateTime(0, 0, tempHours); //show only the hours
    tempHours = changeData(0, 0, tempHours, 2, 0, 23); //modify hours
    updateTime(0, tempMinutes, 0); //show only the minutes
    tempMinutes = changeData(0, tempMinutes, 0, 1, 0, 59); //modify minutes
    updateTime(tempSeconds, 0, 0); //show only the seconds
    tempSeconds = changeData(tempSeconds, 0, 0, 0, 0, 59); //modify seconds
    myOS.pauseTask(blinkClock); //stop LEDs blinking

    //check if the clock was in follow amb light mode
    if (FOLLOW_AMB_LIGHT) {
        myOS.restartTask(readAmbLight);
        myOS.restartTask(writeAmbLight);
    } else {
        analogWrite(LED_POWER, ON);
    }
    //check if the clock was in night mode
    if (NIGHT_MODE && nightModeOn) {
        nightModeOn = false; //so the updateClock routine can put the clock in night mode again
    }
    
    cli(); //stop all the interrupts so I can modify the time variables
    tSeconds = tempSeconds;
    tMinutes = tempMinutes;
    tHours = tempHours;
    sei(); //restart interrupts
    //update local and external clock
    rtc.setTime(tHours, tMinutes, tSeconds);
    updateClock();    
}
    

//change the current value on the display: it can be hours, minutes or seconds
//the data is choosen by the index value, and minVal/maxVal set the range of values
byte changeData(byte hSecs, byte hMins, byte hHrs, byte index, byte minVal, byte maxVal) {
    char timeArray[3] = {hSecs, hMins, hHrs};
    
    waitForRelease(btnSET); //wait for button SET to be released

    unsigned long myMillis = millis(); //read current value of millis
    //programming mode - exit when the user presses the SET button or after 10 seconds (timeout)
    while ((!buttonPressed(btnSET)) && (millis() - myMillis  < 10000)) {
        //if the user presses the INC button, I increment the value
        if (buttonPressed(btnINC)) {
            myMillis = millis();
            if (++timeArray[index] > maxVal) {
                timeArray[index] = minVal;
            }
        }
        //if the user presses the DEC button, I decrement the value
        if (buttonPressed(btnDEC)) {
            myMillis = millis();
            if (--timeArray[index] < minVal) {
                timeArray[index] = maxVal;
            }
        }
        //show the new value
        updateTime(timeArray[0], timeArray[1], timeArray[2]);
        pushData(ledMatrix);
        delay(100);
    }
    return timeArray[index];
}


//read button press (with debounce)
byte buttonPressed(byte buttonPin) {
    if (!digitalRead(buttonPin)) { //has the button been pressed?
        delay(50); //wait for a while
        if (!digitalRead(buttonPin)) { //is the button still pressed?
            delay(50);
            return 1; //yes, it's pressed
        }
    }
    return 0; //NOT pressed, it was a bounce
}


//update the clock - show on the LED matrix the current time
void updateClock() {
    if (NIGHT_MODE) { //is night mode active?
        //if (((tHours * 3600UL) + (tMinutes * 60UL) + tSeconds) < 21600UL) { //night mode is from 00:00:00 to 05:59:59
        if (tHours < 6) {
            if (!nightModeOn) { //the display hasn't been put into night mode yet
                if (FOLLOW_AMB_LIGHT) { //if there's also the follow amb light mode, I disable it
                    myOS.pauseTask(readAmbLight);
                    myOS.pauseTask(writeAmbLight);
                }
                //turn off the display
                analogWrite(LED_POWER, OFF);
                nightModeOn = true;
            }
        } else { //night mode is finished
            if (nightModeOn) { //is the LED matrix still off?
                //push into the matrix the correct time before to light them up
                updateTime(tSeconds, tMinutes, tHours);
                pushData(ledMatrix);
                if (FOLLOW_AMB_LIGHT) { //if there'se also the follow amb light mode, I re-enable it
                    myOS.restartTask(readAmbLight);
                    myOS.restartTask(writeAmbLight);
                } else {
                    analogWrite(LED_POWER, ON);
                }
                nightModeOn = false; //night mode terminated
            }
        }
    }
    
    if (timeChanged && !nightModeOn) { //update the LEDs only if time is changed and it isn't night mode
        timeChanged = false;
        updateTime(tSeconds, tMinutes, tHours);
        pushData(ledMatrix);
    }
}


//prepare the current time for pushing to the led matrix
void updateTime(byte tSecs, byte tMins, byte tHrs) {
    byte i = 0;
    //clear the data
    for (byte j = 0; j < 5; j++) {
        ledMatrix[j] = 0;
    }
    //store units of seconds
    storeData(tSecs % 10, 0);
    //store tens of seconds
    storeData(byte(tSecs / 10), 9);
    //store units of minutes
    storeData(tMins % 10, 14);
    //store tens of minutes
    storeData(byte(tMins / 10), 23);
    //store units of hours
    storeData(tHrs % 10, 28);
    //store tens of hours
    storeData(byte(tHrs / 10), 37);
}


//store data into the matrix
void storeData(byte tempData, byte posData) {
    byte j = posData / 8; //obtain the shift register
    byte i = posData - (j * 8); //obtain the starting pin
    
    while (tempData > 0) {
        bitSet(ledMatrix[j], i);
        if (++i > 7) {
            i = 0;
            j++;
        }
        tempData--;
    }
}


//push data into the LED matrix
void pushData(byte *dataBits) {
    digitalWrite(PIN_LATCH, LOW);
    byte j = 5;
    while (j > 0) {
        shiftOut(PIN_DATA, PIN_CLOCK, MSBFIRST, dataBits[--j]);
    }
    digitalWrite(PIN_LATCH, HIGH);
}




